from glob import glob
import json
import os
from urlparse import urlparse

from twisted.internet.endpoints import serverFromString
from twisted.internet import reactor as default_reactor
from twisted.web.server import Site
from twisted.python import log

from txredisapi import Connection

from .web import PortiaWebServer
from .protocol import JsonProtocolFactory
from .exceptions import PortiaException


def start_redis(redis_uri='redis://localhost:6379/1'):
    try:
        url = urlparse(redis_uri)
    except (AttributeError, TypeError):
        raise PortiaException('Invalid url: %s.' % (redis_uri,))

    if not url.hostname:
        raise PortiaException('Missing Redis hostname.')

    try:
        int(url.path[1:])
    except (IndexError, ValueError):
        raise PortiaException('Invalid Redis db index.')

    return Connection(url.hostname, int(url.port or 6379),
                      dbid=int(url.path[1:]))


def start_webserver(portia, endpoint_str, cors=None, reactor=default_reactor):
    endpoint = serverFromString(reactor, str(endpoint_str))
    return endpoint.listen(
        Site(PortiaWebServer(portia, cors=cors).app.resource()))


def start_tcpserver(portia, endpoint_str, reactor=default_reactor):
    endpoint = serverFromString(reactor, str(endpoint_str))
    return endpoint.listen(JsonProtocolFactory(portia))


def compile_network_prefix_mappings(glob_paths):
    mapping = {}
    for glob_path in glob_paths:
        for mapping_file in glob(glob_path):
            if not os.path.isfile(mapping_file):
                continue

            log.msg('Loading mapping file: %s.' % (mapping_file,))
            with open(mapping_file) as fp:
                mapping.update(json.load(fp))
    return mapping



class PortiaException(Exception):
    pass


class JsonProtocolException(PortiaException):
    def __init__(self, message, command, reference_id):
        super(JsonProtocolException, self).__init__(message)
        self.message = message
        self.command = command
        self.reference_id = reference_id



import json

import dateutil.parser
import phonenumbers

from twisted.internet.protocol import Factory
from twisted.internet.defer import maybeDeferred
from twisted.protocols.basic import LineReceiver

from .exceptions import JsonProtocolException


class JsonProtocol(LineReceiver):

    version = '0.1.0'

    def __init__(self, portia):
        self.portia = portia

    def valid_version(self, received_version):
        return received_version == self.version

    def lineReceived(self, line):
        d = maybeDeferred(self.parseLine, line)
        d.addErrback(self.error)

    def parseLine(self, line):
        data = json.loads(line)
        version = data.get('version')
        command = data.pop('cmd', None)
        reference_id = data.pop('id', None)
        if not self.valid_version(version):
            raise JsonProtocolException(
                'Protocol version mismatch. Expected: %s, got: %s.' % (
                    self.version, version),
                command=command,
                reference_id=reference_id)

        handler = getattr(self, 'handle_%s' % (command,), None)
        if not handler:
            raise JsonProtocolException(
                'Unsupported command: %s.' % (command,),
                command=command,
                reference_id=reference_id)

        d = handler(**data.get('request'))
        d.addCallback(self.reply, command, reference_id)
        d.addErrback(self.error, command, reference_id)
        return d

    def reply(self, data, cmd, reference_id):
        self.sendLine(json.dumps({
            'status': 'ok',
            'cmd': 'reply',
            'reference_cmd': cmd,
            'reference_id': reference_id,
            'version': self.version,
            'response': data,
        }))

    def error(self, failure, command=None, reference_id=None):
        exc = failure.check(JsonProtocolException)
        if exc == JsonProtocolException:
            command = failure.value.command
            reference_id = failure.value.reference_id

        self.sendLine(json.dumps({
            'status': 'error',
            'reference_cmd': command,
            'reference_id': reference_id,
            'message': failure.getErrorMessage(),
            'version': self.version,
        }))

    def handle_get(self, msisdn):
        phonenumber = phonenumbers.parse(msisdn)
        return self.portia.get_annotations(phonenumber)

    def handle_annotate(self, msisdn, key, value, timestamp=None):
        phonenumber = phonenumbers.parse(msisdn)
        if timestamp:
            ts = self.portia.to_utc(dateutil.parser.parse(timestamp))
        else:
            ts = self.portia.now()
        return self.portia.annotate(phonenumber, key, value, timestamp=ts)

    def handle_resolve(self, msisdn):
        return self.portia.resolve(
            phonenumbers.parse(msisdn))


class JsonProtocolFactory(Factory):
    protocol = JsonProtocol

    def __init__(self, portia):
        self.portia = portia

    def buildProtocol(self, *args):
        p = self.protocol(self.portia)
        p.factory = self
        return p



import json
import phonenumbers
from functools import wraps

from twisted.internet import reactor

from klein import Klein

from .exceptions import PortiaException


def validate_key(func):
    @wraps(func)
    def wrapper(portia_server, request, **kwargs):
        try:
            portia_server.portia.validate_annotate_key(kwargs['key'])
        except PortiaException, e:
            request.setResponseCode(400)
            return json.dumps(str(e))

        return func(portia_server, request, **kwargs)
    return wrapper


class PortiaWebServer(object):
    """
    Portia, Number portability as a service
    An API for doing: phone number network lookups.

    :param txredisapi.Connection redis:
        The txredis connection
    """

    app = Klein()
    clock = reactor
    timeout = 5

    def __init__(self, portia, cors=None):
        self.portia = portia
        self.cors = cors

    def default_headers(self, request):
        request.setHeader('Content-Type', 'application/json')
        if self.cors is not None:
            request.setHeader(
                'Access-Control-Allow-Origin',
                self.cors)

    @app.route('/resolve/<msisdn>', methods=['GET'])
    def resolve(self, request, msisdn):
        phonenumber = phonenumbers.parse(msisdn)
        self.default_headers(request)
        d = self.portia.resolve(phonenumber)
        d.addCallback(lambda data: json.dumps(data))
        return d

    @app.route('/entry/<msisdn>', methods=['GET'])
    def get_annotations(self, request, msisdn):
        phonenumber = phonenumbers.parse(msisdn)
        self.default_headers(request)
        d = self.portia.get_annotations(phonenumber)
        d.addCallback(lambda data: json.dumps(data))
        return d

    @app.route('/entry/<msisdn>/<key>', methods=['GET'])
    @validate_key
    def read_annotation(self, request, msisdn, key):
        phonenumber = phonenumbers.parse(msisdn)
        self.default_headers(request)
        d = self.portia.read_annotation(phonenumber, key)
        d.addCallback(lambda data: json.dumps(data))
        return d

    @app.route('/entry/<msisdn>/<key>', methods=['PUT'])
    @validate_key
    def annotate(self, request, msisdn, key):
        phonenumber = phonenumbers.parse(msisdn)
        content = request.content.read()
        self.default_headers(request)

        if not content:
            request.setResponseCode(400)
            return json.dumps('No content supplied')

        d = self.portia.annotate(phonenumber, key, content, self.portia.now())
        d.addCallback(lambda _: json.dumps(content))
        return d
